
/*
 * debtags - Implement package tags support for Debian
 *
 * Copyright (C) 2003  Enrico Zini <enrico@debian.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#define APPNAME PACKAGE_NAME
#else
#warning No config.h found: using fallback values
#define APPNAME __FILE__
#define PACKAGE_VERSION "unknown"
#endif

#include <debtags/Environment.h>
#include <debtags/Consumer.h>
#include <debtags/Package.h>
#include <debtags/PackageSet.h>
#include <debtags/Vocabulary.h>
#include <debtags/TagDB.h>
#include <debtags/MaintainerSet.h>
#include <tagcoll/FilterChain.h>
#include <tagcoll/Implications.h>
#include <tagcoll/TagcollFilter.h>
#include <tagcoll/DerivedTags.h>
#include <tagcoll/InputMerger.h>
#include <tagcoll/TagcollSerializer.h>
#include <tagcoll/TagcollParser.h>
#include <tagcoll/ItemGrouper.h>
#include <tagcoll/ExpressionFilter.h>
#include <tagcoll/TagCollection.h>
#include <tagcoll/TagcollBuilder.h>
#include <tagcoll/Patches.h>
#include <tagcoll/TagcollSink.h>
#include <tagcoll/StdioParserInput.h>
#include <tagcoll/StringParserInput.h>
#include <tagcoll/BasicTagMatcher.h>

#include <tagcoll/Exception.h>
#include <tagcoll/stringf.h>

#include <list>

#include <cctype>

#include <errno.h>
#include <string.h>	// strcasestr

#include "Environment.h"
#include "CommandlineParser.h"

#include <apt-pkg/configuration.h>
#include "acqprogress.h"
#include <termios.h>
#include <sys/ioctl.h>

#include <sys/types.h>  // opendir, readdir, closedir
#include <dirent.h>     // opendir, readdir, closedir
#include <unistd.h>     // access


using namespace std;
using namespace Debtags;

// TODO: remove this, too
const char* fn_tagdb = "/var/lib/debtags/package-tags";
//const char* fn_impls = "/var/lib/debtags/implications";
//const char* fn_deriv = "/var/lib/debtags/derived-tags";
//const char* fn_vocab = "/var/lib/debtags/vocabulary";
//const char* fn_dist_vocab = "/usr/share/debtags/vocabulary";
//const char* fn_sources = "/etc/debtags/sources.list";
//const char* path_cache = "/var/cache/debtags";
//const char* path_tagpatch_d = "/etc/debtags/tagpatch.d";	
//const char* path_tagvoc_d = "/etc/debtags/tagvoc.d";	
//const char* remote_tags_fname = "tags-current.gz";

void wantTagDatabase() throw ()
{
	if (!Debtags::Environment::get().hasTagDatabase())
		fatal_error(
				"The tag database has not yet been generated: you need to run \"debtags update\" (as root) before using the package tags\n");
}

void readCollection(const string& file, TagcollConsumer<string, string>& builder)
	throw (FileException, ParserException)
{
	if (file == "-")
	{
		StdioParserInput input(stdin, "<stdin>");
		Tagcoll::parseTagcoll(input, builder);
	}
	else
	{
		StdioParserInput input(file);
		Tagcoll::parseTagcoll(input, builder);
	}
}

// Catch the tags for a given item
class TagcollCatcher : public TagcollConsumer<string, string>
{
protected:
	string _item;
	OpSet<string> _tags;
	
public:
	TagcollCatcher(const std::string& item) throw ()
		: _item(item) {}
	virtual ~TagcollCatcher() throw () {}

	const OpSet<string>& tags() const throw () { return _tags; }

	// If the package is really untagged, there's nothing we can do here
	virtual void consume(const std::string& item) throw () {}
	virtual void consume(const OpSet<string>& items) throw () {}
	
	// Catch the tags that belong to _item
	virtual void consume(const std::string& item, const OpSet<string>& tags) throw ()
	{
		if (item == _item)
			_tags += tags;
	}

	// Catch the tags that belong to _item
	virtual void consume(const OpSet<string>& items, const OpSet<string>& tags) throw ()
	{
		if (items.contains(_item))
			_tags += tags;
	}
};

// Check a collection agaisnt a tag vocabulary
class TagcollChecker : public TagcollConsumer<string>
{
protected:
	const Debtags::Vocabulary& voc;
	int missing_count;
	map<string, int> missing;
	unsigned int mtag_width;
	int max_missing;

	void checkTag(const std::string& tag) throw ()
	{
		if (!voc.hasTag(tag))
		{
			//warning("Tag %.*s not found in vocabulary\n", PFSTR(tag));
			missing_count++;
			missing[tag]++;
			if (tag.size() > mtag_width)
				mtag_width = tag.size();
			if (missing[tag] > max_missing)
				max_missing = missing[tag];
		}
	}
	
public:
	TagcollChecker(const Debtags::Vocabulary& voc) throw ()
		: voc(voc), missing_count(0), mtag_width(0), max_missing(0) {}
	virtual ~TagcollChecker() throw () {}

	// Get the number of times a bad tag has been found in input
	int missingCount() const throw () { return missing_count; }

	// If the package is really untagged, there's nothing we can do here
	virtual void consume(const std::string& item) throw () {}
	virtual void consume(const OpSet<string>& items) throw () {}
	
	// Catch the tags that belong to _item
	virtual void consume(const std::string& item, const OpSet<string>& tags) throw ()
	{
		for (OpSet<string>::const_iterator i = tags.begin(); i != tags.end(); i++)
			checkTag(*i);
	}

	// Catch the tags that belong to _item
	virtual void consume(const OpSet<string>& items, const OpSet<string>& tags) throw ()
	{
		for (OpSet<string>::const_iterator i = tags.begin(); i != tags.end(); i++)
			checkTag(*i);
	}

	void report() throw ()
	{
		if (missing_count > 0)
		{
			warning("%d missing tags found in %d occasions\n", missing.size(), missing_count);
			warning("Missing tags were:\n");

			int mtag_maxdigits = 0;
			for (int i = max_missing; i > 0; i = i / 10)
				mtag_maxdigits++;

			for (map<string, int>::const_iterator i = missing.begin();
					i != missing.end(); i++)
			{
				warning("\t%.*s %*s (%*d time%s)\n",
					PFSTR(i->first), 
					mtag_width - i->first.size(),
					"", 
					mtag_maxdigits,
					i->second,
					i->second > 1 ? "s" : "");
			}
		}
	}
};

class ItemCollector : public OpSet<string>, public TagcollConsumer<string, string>
{
public:
	virtual ~ItemCollector() throw() {}

	// Process an untagged item
	virtual void consume(const string& item) throw () { (*this) += item; }

	// Process a tagged item, with its tags
	virtual void consume(const string& item, const OpSet<string>& tags) throw ()  { (*this) += item; }

	// Process a set of items, all with no tags
	virtual void consume(const OpSet<string>& items) throw () { (*this) += items; }

	// Process a set of items identically tagged, with their tags
	virtual void consume(const OpSet<string>& items, const OpSet<string>& tags) throw ()
		{ (*this) += items; }
};

class PackagePrinter : public Debtags::PackageConsumer
{
protected:
	int _count;
	Debtags::PackageSet blacklist;

public:
	PackagePrinter() throw () : _count(0) {}
	PackagePrinter(const Debtags::PackageSet& blacklist) throw ()
		: _count(0), blacklist(blacklist) {}
	virtual ~PackagePrinter() throw () {}

	virtual void consume(const Debtags::Package& pkg) throw ()
	{
		if (blacklist.find(pkg) == blacklist.end())
		{
			printf("%.*s - %.*s\n", PFSTR(pkg.name()), PFSTR(pkg.sdesc()));
			_count++;
		}
	}
	int count() const throw () { return _count; }
};

static void printShortVocabularyItem(const Tag& tag)
{
	printf("%.*s - %.*s\n", PFSTR(tag.fullname()), PFSTR(tag.sdesc()));
}

static void printVocabularyItem(const Tag& tag)
{
	string buf = tag.fulldata();
	
	// Count the end of lines at the end of the buffer
	int nnl = 0;
	while (buf[buf.size() - 1 - nnl] == '\n')
		nnl++;
	fwrite(buf.data(), buf.size() - nnl, 1, stdout);
	fputc('\n', stdout);
	fputc('\n', stdout);
}

static bool matchVocabularyItem(const Tag& tag, const string& pat) throw ()
{
	if (strcasestr(tag.name().c_str(), pat.c_str())
		|| strcasestr(tag.sdesc().c_str(), pat.c_str())
		|| strcasestr(tag.ldesc().c_str(), pat.c_str()))
		return true;
	/* TODO
	else
		for (list< pair<string, string> >::const_iterator j = rec.others.begin();
				j != rec.others.end(); j++)
			if (strcasestr(j->second.c_str(), pat.c_str()))
				return true;
	*/
	return false;
/*
	if (tag.find(pat) != string::npos
		|| rec.description.find(pat) != string::npos)
		return true;
	else
		for (list< pair<string, string> >::const_iterator j = rec.others.begin();
				j != rec.others.end(); j++)
			if (j->second.find(pat) != string::npos)
				return true;
	return false;
*/
}

class CommandlineParserWithCommand : public CommandlineParser
{
protected:
	map<string, int> command_map;

public:
	CommandlineParserWithCommand(const std::string& argv0,
									const std::string& cmdline_summary,
									const std::string& description) throw ()
		: CommandlineParser(argv0, cmdline_summary, description)
	{
		add("version", 'V', "version", "print the program version, then exit");
	}

	void addCommand(const std::string name, int id) throw ()
	{
		command_map[name] = id;
	}

	int parse(int& argc, const char**& argv) throw ()
	{
		if (!CommandlineParser::parse(argc, argv))
		{
			printHelp();
			exit(1);
		}
		if (get("help").defined())
		{
			printHelp();
			exit(0);
		}
		if (get("version").defined())
		{
			printf("%s ver." PACKAGE_VERSION "\n", APPNAME);
			exit(0);
		}
		if (argc == 1)
		{
			printHelp();
			exit(1);
		}

		string command_string = argv[1];
		map<string, int>::const_iterator cmap_i = command_map.find(command_string);
		if (cmap_i == command_map.end())
		{
			fprintf(stderr, "Invalid command: \"%.*s\"\n", PFSTR(command_string));
			printHelp();
			exit(1);
		}

		for (int i = 1; i < argc; i++)
			argv[i] = argv[i + 1];
		--argc;
		
		return cmap_i->second;
	}
};

class CommandlineArgs
{
protected:
	int argc;
	const char** argv;
	int _next;

public:
	CommandlineArgs(int argc, const char* argv[]) throw () : argc(argc), argv(argv), _next(1) {}

	// Return true if there is another argument left in the list
	bool hasNext() const throw () { return argc >= _next + 1; }

	// Return the next argument in the list
	string next() throw ()
	{
		if (hasNext())
		{
			return argv[_next++];
		} else {
			return "-";
		}
	}
};

enum valid_command { UPDATE, CHECK, TAGSHOW, TAGSEARCH, SHOW, RELATED, CAT, GREP, INSTALL, MKPATCH, MAINTAINERS };

int main(int argc, const char* argv[])
{
	try {
		// Install the handler for unexpected exceptions
		InstallUnexpected installUnexpected;

		CommandlineParserWithCommand opts(APPNAME, "[options] [command] [file1|-] [file2|-]",
				"Perform various operations on a tagged collection\n\n"
				"Commands are:\n"
				"  update        Updates the package tag database (requires root)\n"
				"  check <file>	 Check that all the tags in the given tagged collection are present\n"
				"                in the tag vocabulary.  Checks the main database if no file is\n"
				"                specified\n"
				"  tagshow <tag> Show the vocabulary informations about a tag\n"
				"  tagsearch <string [string [string [...]]]>\n"
				"                Show a summary of all tags whose data contains the given strings\n"
				"  show <pkg>    Call apt-cache show <pkg>, but add tag informations to the output.\n"
				"  related <pkg1[,pkg2[,pkg2,[...]]]>\n"
				"                Show packages related to the specified ones.\n"
				"  cat           Output the full package tag database\n"
				"  grep [-v] [-q] <tag expression>\n"
				"                Output the lines of the full package tag database that match\n"
				"                the given tag expression\n"
				"  install [-v] [-q] <tag expression>\n"
				"                apt-get install the packages that match the given tag expression\n"
				"  mkpatch [filename]\n"
				"                Create a tag patch between the current tag database and the tag\n"
				"                collection [filename]\n"
				"  maintainers   Create a tagged collection of maintainers and the tags of the\n"
				"                packages they maintain\n");


		/*
		opts.add("hierarchy", 's', "smart-hierarchy", "build a smart hierarchy");
		opts.add("implications", 'm', "show-implications", "output a list of tag implications");
		opts.add("copy", 'c', "copy", "output the collection");
		opts.add("diff", 'd', "diff", "output a tag patch file with the differences between two files (requires two file arguments)");
		*/

		//opts.add("expanded", 'x', "expanded-output", "produce full (and redundant) output data instead of compact");
		opts.add("groupitems", 'g', "group-items", "group items with the same tagset in the output collection");
		opts.add("distance", 'd', "distance", "set the maximum distance to use for the \"related\" command (defaults to 0)", "num");

		opts.add("invert-match", 'v', "invert-match", "invert the sense of matching, to select non-matching lines");
		opts.add("quiet", 'q', "quiet", "do not write anything to standard output, but exit with 0 if any match is found");
		
		opts.add("verbose", 0, "verbose", "enable verbose output");
		opts.add("debug", 0, "debug", "enable debugging output (including verbose output)");

		opts.addCommand("update", (int)UPDATE);
		opts.addCommand("check", (int)CHECK);
		opts.addCommand("tagshow", (int)TAGSHOW);
		opts.addCommand("tagsearch", (int)TAGSEARCH);
		opts.addCommand("show", (int)SHOW);
		opts.addCommand("related", (int)RELATED);
		opts.addCommand("cat", (int)CAT);
		opts.addCommand("grep", (int)GREP);
		opts.addCommand("install", (int)INSTALL);
		opts.addCommand("mkpatch", (int)MKPATCH);
		opts.addCommand("maintainers", (int)MAINTAINERS);

		// Process the commandline
		valid_command cmd = (valid_command)opts.parse(argc, argv);

		CommandlineArgs args(argc, argv);

		if (opts.get("verbose").defined())
			::Environment::get().verbose(true);

		if (opts.get("debug").defined())
			::Environment::get().debug(true);

		Debtags::Environment::init(false);
		
		// Perform the correct operation
		switch (cmd)
		{
			// Output the full package tag database
			case CAT:
			{
				wantTagDatabase();
				TagcollSerializer serializer(stdout);
				if (opts.get("groupitems").defined())
				{
					ItemGrouper<string, string> grouper;
					TagDB::outputPatched(grouper);
					grouper.output(serializer);
				} else
					TagDB::outputPatched(serializer);

				break;
			}

			// grep [-v] [-q] <tag expression>
			// Output the lines of the full package tag database that match the
			// given tag expression
			case GREP:
			{
				wantTagDatabase();
				string expression = args.next();

				// Build the grep filter chain
				FilterChain<string, string> filters;
				ExpressionFilter filter;
				if (!filter.setExpression(expression))
					fatal_error("Error in tag expression `%.*s'.\n", PFSTR(expression));
				filters.appendFilter(&filter);

				if (opts.get("invert-match").defined())
					filter.setMatchType(ExpressionFilter::INVERTED);

				if (opts.get("quiet").defined())
					filters.setConsumer(new TagcollSink<string>());
				else
					filters.setConsumer(new TagcollSerializer(stdout));

				Debtags::Environment::get().tagDB().outputPatched(filters);

				return filter.countMatched() > 0 ? 0 : 1;
			}

			// install [-v] [-q] <tag expression>
			// apt-get install the packages that match the given tag expression
			case INSTALL:
			{
				wantTagDatabase();
				string expression = args.next();

				// Build the grep filter chain
				FilterChain<string, string> filters;
				ExpressionFilter filter;
				if (!filter.setExpression(expression))
					fatal_error("Error in tag expression `%.*s'.\n", PFSTR(expression));
				filters.appendFilter(&filter);

				if (opts.get("invert-match").defined())
					filter.setMatchType(ExpressionFilter::INVERTED);

				ItemCollector selection;
				filter.setConsumer(&selection);
				Debtags::Environment::get().tagDB().outputPatched(filters);

				const char* parms[selection.size() + 1];
				int i = 0;
				parms[i++] = APTGET;
				parms[i++] = "install";
				for (ItemCollector::iterator tag = selection.begin();
						tag != selection.end(); tag++)
					parms[i++] = tag->c_str();
				parms[i] = 0;
				execv(APTGET, (char* const*)parms);
				throw SystemException(errno, "Running apt-get");
			}

			// tagshow <tag>
			// Show the vocabulary informations about a tag\n"
			case TAGSHOW:
			{
				wantTagDatabase();
				string tag = args.next();

				Debtags::Vocabulary& voc = Debtags::Environment::get().vocabulary();
				Tag t = voc.findTag(tag);
				if (!t)
				{
					verbose("Tag `%.*s' was not found in tag vocabulary\n", PFSTR(tag));
					return 1;
				}
				else
				{
					printVocabularyItem(t);
					return 0;
				}
			}
			// tagsearch <pattern [pattern [pattern [...]]]>
			// Show a summary of all tags matching the given patterns
			case TAGSEARCH:
			{
				wantTagDatabase();
				// Get the patterns to be matched
				vector<BasicTagMatcher> matchers;
				while (args.hasNext())
				{
					BasicTagMatcher tm;
					tm.setString(args.next());
					matchers.push_back(tm);
				}

				if (matchers.empty())
				{
					error("No patterns provided in commandline\n");
					return 1;
				}

				Debtags::Vocabulary& voc = Debtags::Environment::get().vocabulary();

				TagSet matches = voc.getFiltered(matchers[0]);
				for (unsigned int i = 1; i < matchers.size(); i++)
					matches = matches.getFiltered(matchers[i]);
				
				for (TagSet::const_iterator i = matches.begin();
						i != matches.end(); i++)
					printShortVocabularyItem(*i);

				return matches.size() > 0 ? 0 : 1;
			}
			// show <pkg>
			// Call apt-cache show <pkg>, but add tag informations to the output.\n"
			case SHOW:
			{
				using namespace Debtags;
				
				wantTagDatabase();

				string name = args.next();
				
				Package pkg = Debtags::Environment::get().packages().find(name);
				if (pkg)
				{
					printf("%.*sTags: ", PFSTR(pkg.fulldata()));
					TagSet ts = pkg.tags();
					for (TagSet::const_iterator i = ts.begin();
							i != ts.end(); i++)
						if (i == ts.begin())
							printf("%.*s", PFSTR(i->fullname()));
						else
							printf(", %.*s", PFSTR(i->fullname()));
					printf("\n");
					return 0;
				} else {
					verbose("Package %.*s not found", PFSTR(name));
					return 1;
				}
			}
			// related <pkg1[,pkg2[,pkg2,[...]]]>
			// Show packages related to the specified ones
			case RELATED:
			{
				using namespace Debtags;

				wantTagDatabase();

				int maxdist = 0;
				if (opts.get("distance").defined())
					maxdist = opts.get("distance").intVal();
				string pkg = args.next();

				const Debtags::Environment& env = Debtags::Environment::get();

				// Split the items on commas
				string splititem;
				PackageSet splititems;
				for (string::const_iterator c = pkg.begin(); c != pkg.end(); c++)
					if (*c == ',')
					{
						if (Package p = env.getPackage(splititem))
						{
							splititems.insert(p);
						} else {
							error("Item \"%.*s\" does not exist in the collection\n", PFSTR(splititem));
							return 1;
						}
						splititem = string();
					} else
						splititem += *c;
				if (Package p = env.getPackage(splititem))
				{
					splititems.insert(p);
				} else {
					error("Item \"%.*s\" does not exist in the collection\n", PFSTR(splititem));
					return 1;
				}

				// Get the tagset as the intersection of the tagsets of all input items
				PackageSet::const_iterator i = splititems.begin();
				TagSet ts = (*i).tags();
				for (++i; i != splititems.end(); i++)
					ts = ts ^ (*i).tags();

				if (ts.empty())
				{
					if (splititems.size() > 1)
						fprintf(stderr, "The packages %.*s are unrelated: cannot find a barycenter to start computing relationships from.\n", PFSTR(pkg));
					else
						fprintf(stderr, "The package %.*s has no tags attached.\n", PFSTR(pkg));
					return 1;
				}

				PackagePrinter printer(splititems);
				Debtags::Environment::get().outputRelated(printer, ts, maxdist);

				if (printer.count() > 50 && maxdist == 0 && isatty(1))
				{
					string tags;
					for (TagSet::const_iterator i = ts.begin();
							i != ts.end(); i++)
						if (i == ts.begin())
							tags += i->fullname();
						else
							tags += "%2C" + i->fullname();
					feedback("\nIt seems that this set of packages lacks tag information that could help to better distinguish package similarities.\nYou can help providing better tagging: just point your browser to http://debian.vitavonni.de/packagebrowser/?tags=%.*s\n", PFSTR(tags));
				}
				return 0;
			}
			// check <file>
			// Check that all the tags in the given tagged collection are
			// present in the tag vocabulary.  Checks the main database if no
			// file is specified
			case CHECK:
			{
				wantTagDatabase();
				string file;
				if (args.hasNext())
					file = args.next();
				else
					file = fn_tagdb;

				const Debtags::Vocabulary& voc = Debtags::Environment::get().vocabulary();

				TagcollChecker checker(voc);
				readCollection(file, checker);

				if (checker.missingCount() > 0)
				{
					checker.report();
					return 1;
				}
				else
					return 0;
			}
			// update
			// Updates the package tag database (requires root)
			case UPDATE:
			{
				if (geteuid() != 0)
				{
					throw ConsistencyCheckException("You must be root to update the system debtags database");
				}
				struct winsize ws;
				unsigned int ScreenWidth;

				if (ioctl(1, TIOCGWINSZ, &ws) != -1 && ws.ws_col >= 5)
					ScreenWidth = ws.ws_col - 1;

				AcqTextStatus Stat(ScreenWidth,_config->FindI("quiet",0));
				Debtags::Environment::get().updateDebtagsDatabase(&Stat);
				feedback("Done.\n");
				break;
			}
			// mkpatch [filename]
			// Create a tag patch between the current tag database and the tag
			// collection [filename]
			case MKPATCH:
			{
				wantTagDatabase();
				string file = args.next();

				InputMerger<string,string> merger1;
				InputMerger<string,string> merger2;

				Debtags::Environment::get().tagDB().outputPatched(merger1);
				Debtags::Environment::get().tagDB().outputFile(file, merger2);

				PatchList<string> newpatches;
				newpatches.generate(merger1, merger2);

				TagcollSerializer output(stdout);
				newpatches.output(output);
				break;
			}

			// maintainers
			// Create a tagged collection of maintainers and the tags of the
			// packages they maintain
			case MAINTAINERS:
			{
				using namespace Debtags;

				wantTagDatabase();

				MaintainerSet maints = Debtags::Environment::get().maintainers();
				
				for (MaintainerSet::const_iterator i = maints.begin();
						i != maints.end(); i++)
				{
					printf("%.*s: ", PFSTR(i->fullname()));

					const PackageSet& packages = i->packages();
					TagSet tags;
					for (PackageSet::const_iterator j = packages.begin();
							j != packages.end(); j++)
						tags += j->tags();

					for (TagSet::const_iterator j = tags.begin();
							j != tags.end(); j++)
						if (j != tags.begin())
							printf(", %.*s", PFSTR(j->fullname()));
						else
							printf("%.*s", PFSTR(j->fullname()));

					printf("\n");
				}
				break;
			}
		}

		return 0;
	}
	catch (Exception& e)
	{
		fatal_error("%s: %.*s\n", e.type(), PFSTR(e.desc()));
		return 1;
	}
}

// vim:set ts=4 sw=4:
